import json
import os
import sys
from datetime import datetime
from typing import Optional

import click


class Directory:
    client = None
    upload_folder = None
    counters = None

    def __init__(  # noqa
            self,
            strftime_format: str,
            name: Optional[str],
            capacity: Optional[int],
            counter_max: Optional[int],
    ):
        self._strftime_format = strftime_format
        self.name = name
        self.capacity = capacity
        self.counter_max = counter_max

    def generate_name(self, file_upload_date: Optional[datetime] = None):
        if file_upload_date is None:
            file_upload_date = datetime.now()
        return file_upload_date.strftime(self._strftime_format)


class LastDirectory(Directory):
    def __init__(self, strftime_format: str, name: Optional[str] = None):
        super().__init__(strftime_format, name, None, None)


class Counters:

    def __init__(self, app_name):
        self.updates = False

        self._file_name = f"/var/lib/backup_uploader/{app_name}.json"

        try:
            with open(self._file_name) as counters_file:
                try:
                    self._json = json.load(counters_file)
                except json.decoder.JSONDecodeError:
                    click.echo(f'Counters file "{self._file_name}" with invalid json format', err=True)
                    sys.exit(2)
        except FileNotFoundError:
            self._json = {}
            self.updates = True
        except PermissionError:
            click.echo(f'Unable to read on counters file "{self._file_name}"', err=True)
            sys.exit(2)

        old_mask = os.umask(0o002)
        try:
            with open(self._file_name, "a+") as f:  # test if I can write/create the file to later save
                if f.tell() == 0:
                    f.write("{}")
        except FileNotFoundError:
            click.echo(
                "The directory /var/lib/backup_uploader does not exist.\n"
                'Run sudo sh -c "useradd backup_uploader && '
                'mkdir /var/lib/backup_uploader && '
                'chown backup_uploader:backup_uploader /var/lib/backup_uploader && '
                'chmod 775 /var/lib/backup_uploader" '
                'to create it and the respective backup_uploader user and group.'
            )
        except PermissionError:
            click.echo(
                f'Unable to create/write on counters file "{self._file_name}".\n'
                'The user running this script should belong to the "backup_uploader" group. '
                'Run sudo sh -c "usermod -a -G backup_uploader $USER".',
                err=True,
            )
            sys.exit(2)
        finally:
            os.umask(old_mask)

    def inc(self, name):
        if name not in self._json:
            self._json[name] = 0
        self._json[name] += 1
        self.updates = True
        return self._json[name]

    def reset(self, name):
        self._json[name] = 0
        self.updates = True

    def get(self, name):
        return self._json[name]

    def save(self):
        if self.updates:
            old_mask = os.umask(0o002)
            try:
                with open(self._file_name, "w+") as counters_file:
                    json.dump(self._json, counters_file)
            finally:
                os.umask(old_mask)



import datetime
import sys

import mega
import mega.errors

from backup_uploader.clients.base import BaseClient
from backup_uploader.clients import exceptions
from backup_uploader.directories import Directory


class MegaClient(BaseClient):

    def __init__(self, app_name):
        super(MegaClient, self).__init__(app_name)
        self._client = mega.Mega()

    def login(self, credentials_file):
        try:
            email = credentials_file.readline().strip()
            password = credentials_file.readline().strip()
        except IOError:
            raise exceptions.InvalidCredentialsFileFormat()

        try:
            self._client.login(email, password)
        except mega.errors.RequestError as err:
            raise exceptions.UnableToLogin(err)

    def get_or_create_directory(self, name: str):
        if self._root_dir is not None:
            name = f"{self._app_name}/{name}"

        directory = self._client.find(name, exclude_deleted=True)
        if not directory:
            return self._client.create_folder(name, dest=self._root_dir)

        if directory[1]["t"] != 1:  # if its not a directory
            print(
                f"Node on the {name} directory provided is not a directory.",
                file=sys.stderr,
            )
            sys.exit(3)

        return directory[0]

    def get_files_in_directory(self, directory: Directory):
        if not hasattr(directory, "id"):
            return self.get_or_create_directory(directory.name)

        return self._client.get_files_in_node(directory.id)

    def get_oldest_file(self, files):
        return sorted(files.items(), key=lambda f: f[1]["ts"])[0]

    def move_file(self, file, directory: Directory):
        if not hasattr(directory, "id"):
            directory.id = self.get_or_create_directory(directory.name)

            if isinstance(directory.id, dict):
                directory.id = directory.id[directory.name]

        self._client.move(file, directory.id)
        self._client.rename(
            file, directory.generate_name(datetime.datetime.fromtimestamp(file[1]["ts"]))
        )

    def upload_file(self, filename, directory: Directory):
        dest = self.get_or_create_directory(directory.name)
        directory.id = dest

        file = self._client.upload(filename, dest, directory.generate_name())["f"][0]["h"]
        file = self._client.find(handle=file)
        self._client.rename((None, file), directory.generate_name())

    def delete_file(self, file):
        self._client.destroy(file[1])



from typing import Optional

from backup_uploader.clients.base import BaseClient
from backup_uploader.directories import Counters, Directory, LastDirectory


class BackupChain:

    def __init__(self, client: BaseClient, counters: Counters):
        self._directories = []
        self._last = None

        self._client = client
        self._counters = counters

    def add_directory(self, strftime_format: str, name: str, capacity: int, counter_max: int):
        self._directories.append(Directory(strftime_format, name, capacity, counter_max))
        return self

    def set_last(self, strftime_format: str, name: Optional[str] = None):
        self._last = LastDirectory(strftime_format, name)

    def _first(self):
        if self._directories:
            return self._directories[0]
        return self._last

    def empty(self):
        return len(self._directories) == 0

    def store(self, filename):
        self._client.upload_file(filename, self._first())

        for i, directory in enumerate(self._directories):
            if i != 0:
                self._client.move_file(file, directory)

            stored_files = self._client.get_files_in_directory(directory)
            if len(stored_files) <= directory.capacity:
                return

            oldest_file = self._client.get_oldest_file(stored_files)

            if self._counters.inc(directory.name) >= directory.counter_max:
                file = oldest_file
                self._counters.reset(directory.name)
            else:
                self._client.delete_file(oldest_file)
                return

        if self._last is not None:
            if self._directories:
                self._client.move_file(file, self._last)
            # if there is no intermediary directories the file was already written to the last directory
        else:
            self._client.delete_file(file)



import importlib
import importlib.util
import sys

import click

from backup_uploader.chain import BackupChain
from backup_uploader.clients.exceptions import BaseClientException
from backup_uploader.directories import Counters
from backup_uploader.parser import InvalidChainConfig, parse_chain_config_file


@click.command()
@click.argument(
    "app_name",
    type=click.STRING,
    required=True,
)
@click.argument(
    "client",
    type=click.STRING,
    required=True,
)
@click.argument(
    "credentials_file",
    type=click.File("r"),
    required=True,
)
@click.argument(
    "chain_config",
    type=click.STRING,
    required=True,
)
@click.argument(
    "upload_file",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
    ),
    required=True,
)
def main(
    app_name,
    client,
    credentials_file,
    chain_config,
    upload_file,
):
    counters = Counters(app_name)

    client = client.lower()

    if not importlib.util.find_spec(f"backup_uploader.clients.{client}"):
        click.echo(f"Unknown client {client}", err=True)
        return 1

    try:
        client_module = importlib.import_module(f"backup_uploader.clients.{client}")
    except ImportError:
        click.echo(f"Missing python package for client {client}", err=True)
        return 1
    client = getattr(client_module, f"{client.title()}Client")(app_name)

    try:
        client.login(credentials_file)
    except BaseClientException as ex:
        click.echo(ex.message, err=True)
        return 1

    chain = BackupChain(client, counters)

    try:
        parse_chain_config_file(chain_config, chain)
    except InvalidChainConfig as ex:
        click.echo(ex.message, err=True)
        return 1

    with client:
        chain.store(upload_file)

    counters.save()


if __name__ == '__main__':
    main(sys.argv[1:])



from datetime import datetime

from backup_uploader.chain import BackupChain

DIRECTORY_STRFTIME_FORMAT = 1
DIRECTORY_NAME = 0
DIRECTORY_CAPACITY = 2
DIRECTORY_COUNTER_MAX = 3


def parse_chain_config_file(config_text, chain: BackupChain):
    used_names = set()
    duplicated_names = set()
    exceptions = []

    config = config_text.split(";")

    if len(config) == 0:
        raise InvalidChainConfig((EmptyConfig(),))

    for position, directory in enumerate(config):
        directory = directory.split(":")

        if len(directory) != 2 and len(directory) != 4:
            raise InvalidFieldCombination(position)

        name = directory[DIRECTORY_NAME]
        if name in used_names:
            duplicated_names.add(name)
            continue

        used_names.add(name)

        strftime_format = directory[DIRECTORY_STRFTIME_FORMAT]
        try:
            datetime.now().strftime(strftime_format)
        except ValueError:
            exceptions.append(InvalidStrftimeFormat(strftime_format, position))

        if len(directory) == 4:
            try:
                capacity = int(directory[DIRECTORY_CAPACITY])
                counter_max = int(directory[DIRECTORY_COUNTER_MAX])
            except ValueError:
                exceptions.append(InvalidIntegerFormat(position))
            else:
                chain.add_directory(
                    strftime_format,
                    name,
                    capacity,
                    counter_max,
                )
        else:
            chain.set_last(strftime_format, name)
            break

    if duplicated_names:
        exceptions.append(DuplicatedDirectoryName(duplicated_names))

    if exceptions:
        raise InvalidChainConfig(exceptions)
    elif position != len(config) - 1:
        raise InvalidChainConfig((DirectoriesAfterLast(position),))


class InvalidChainConfig(Exception):
    def __init__(self, exceptions):
        self.message = '; '.join(ex.message for ex in exceptions)


class EmptyConfig(InvalidChainConfig):
    def __init__(self):
        self.message = "Empty chain config was provided"
        super(InvalidChainConfig, self).__init__()


class MissingMandatoryField(InvalidChainConfig):
    def __init__(self, missing_keys, directory_position):
        self.message = f'Missing mandatory key(s) {", ".join(missing_keys)} on directory in position {directory_position}'
        super(InvalidChainConfig, self).__init__()


class DuplicatedDirectoryName(InvalidChainConfig):
    def __init__(self, duplicated_names):
        self.message = f"The directory name(s) {', '.join(duplicated_names)} are duplicated"
        super(InvalidChainConfig, self).__init__()


class InvalidStrftimeFormat(InvalidChainConfig):
    def __init__(self, format, directory_position):
        self.message = f'Invalid strftime format "{format}" on directory in position {directory_position}'
        super(InvalidChainConfig, self).__init__()


class InvalidIntegerFormat(InvalidChainConfig):
    def __init__(self, directory_position):
        self.message = f'Invalid integers on key(s) {DIRECTORY_CAPACITY} and/or ' \
                       f'{DIRECTORY_COUNTER_MAX} on directory in position {directory_position}'
        super(InvalidChainConfig, self).__init__()


class InvalidFieldCombination(InvalidChainConfig):
    def __init__(self, directory_position):
        self.message = f"Directory on position {directory_position} must have either 2 (last directory) " \
                       f"or 4 (intermediate directory) fields"
        super(InvalidChainConfig, self).__init__()


class DirectoriesAfterLast(InvalidChainConfig):
    def __init__(self, directory_position):
        self.message = f"There are directories defined after a LastDirectory, defined on position {directory_position}"
        super(InvalidChainConfig, self).__init__()



from backup_uploader.clients.base import BaseClient
from backup_uploader.clients.exceptions import UnableToLogin
from backup_uploader.directories import Directory

import dropbox
import dropbox.exceptions
import dropbox.files


class DropboxClient(BaseClient):

    CHUNK_SIZE = 100 * 1024 * 1024  # 100 MB

    def __init__(self, app_name):
        super(DropboxClient, self).__init__(app_name)
        # self._client = dropbox.Dropbox()

    def login(self, credentials_file):
        self._client = dropbox.Dropbox(credentials_file.readline().strip())
        self._client.__enter__()

        # Check that the access token is valid
        try:
            self._client.users_get_current_account()
        except dropbox.exceptions.AuthError:
            raise UnableToLogin(
                "ERROR: Invalid access token; try re-generating an "
                "access token from the app console on the web."
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._client.__exit__()

    def get_or_create_directory(self, name: str):
        if self._root_dir is not None:
            name = f"/{self._app_name}/{name}"
        else:
            name = f"/{name}"

        try:
            self._client.files_create_folder_v2(name)
        except dropbox.exceptions.ApiError:
            pass  # already exists

        return name

    def get_files_in_directory(self, directory: Directory):
        name = self.get_or_create_directory(directory.name)

        return self._client.files_list_folder(name).entries

    def get_oldest_file(self, files):
        files.sort(key=lambda f: f.client_modified)
        return files[0]

    def move_file(self, file, directory: Directory):
        dir_name = self.get_or_create_directory(directory.name)

        self._client.files_move_v2(
            file.path_display,
            f"{dir_name}/{directory.generate_name(file.client_modified)}",
            autorename=True,
        )

    def upload_file(self, filename: str, directory: Directory):
        dir_name = self.get_or_create_directory(directory.name)

        with open(filename, "rb") as file:
            data = file.read(self.CHUNK_SIZE)
            upload_session_result = self._client.files_upload_session_start(data)
            cursor = dropbox.files.UploadSessionCursor(
                upload_session_result.session_id,
                file.tell(),
            )

            while True:
                cursor.offset = file.tell()

                data = file.read(self.CHUNK_SIZE)

                if not data:
                    break

                self._client.files_upload_session_append_v2(
                    data,
                    cursor,
                )

            self._client.files_upload_session_finish(
                b"",
                cursor,
                dropbox.files.CommitInfo(
                    f"{dir_name}/{directory.generate_name()}",
                    autorename=True,
                )
            )

    def delete_file(self, file):
        self._client.files_delete_v2(file.path_display)



from abc import ABC, abstractmethod

from backup_uploader.directories import Directory


class BaseClient(ABC):

    def __init__(self, app_name):
        self._app_name = app_name
        self._root_dir = None

    @abstractmethod
    def login(self, credentials_file):
        pass

    @abstractmethod
    def get_or_create_directory(self, name: str):
        pass

    @abstractmethod
    def get_files_in_directory(self, directory: Directory):
        pass

    @abstractmethod
    def get_oldest_file(self, files):
        pass

    @abstractmethod
    def move_file(self, file, directory: Directory):
        pass

    @abstractmethod
    def upload_file(self, filename: str, directory: Directory):
        pass

    @abstractmethod
    def delete_file(self, file):
        pass

    def __enter__(self):
        self._root_dir = self.get_or_create_directory(self._app_name)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass




class BaseClientException(Exception):
    def __init__(self, message):
        self.message = message
        super(BaseClientException, self).__init__()


class InvalidCredentialsFileFormat(BaseClientException):
    def __init__(self, message):
        super(InvalidCredentialsFileFormat, self).__init__(message)


class UnableToLogin(BaseClientException):
    def __init__(self, message):
        super(UnableToLogin, self).__init__(message)



from setuptools import setup, find_packages
import json
import os

with open(os.path.join(os.path.dirname(__file__), "version.json")) as version_file:
    version = json.load(version_file)["version"]

setup(
    name="backup_uploader",
    packages=find_packages(include=[
        "backup_uploader",
        "backup_uploader.clients",
    ]),
    license="MIT",
    version=version,
    url="https://github.com/aspedrosa/BackupsUploader",
    author="André Pedrosa",
    author_email="aspedr0sa@protonmail.com",
    keywords=["backup", "uploader"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
    include_package_data=True,
    install_requires=["click"],
    extras_require={
        "dropbox": ["dropbox"],
        "mega": ["mega.py"],
    },
    entry_points="""
    [console_scripts]
    backup_uploader=backup_uploader.cli:main
    """,
)
